/*$Id: d_ccvs.cc,v 13.40 1999/05/19 08:40:21 al Exp $ -*- C++ -*-
 * functions for ccvs
 * It is really voltage controlled, taking into account the sense element.
 * Then adjust the gain to account for the sense element.
 */
#include "d_ccvs.h"
/*--------------------------------------------------------------------------*/
//	void	DEV_CCVS::expand();
// 	bool	DEV_CCVS::dotr();
// 	void	DEV_CCVS::doac();
/*--------------------------------------------------------------------------*/
CARD& DEV_CCVS::expand()
{
  if (*inputlabel){
    input = dynamic_cast<ELEMENT*>(find_in_scope(inputlabel));
  }else{
    // input already set, an internal element.  none used yet
    untested();
  }
  if (!input){
    untested();
    error(bERROR,"%s: %s cannot be used as current probe\n",
	  printlabel(),inputlabel);
  }
  if (input->is2port()){
    untested();
    n[IN1] = input->n[IN1];
    n[IN2] = input->n[IN2];
  }else{
    assert(input->is1port()  ||  input->issource());
    n[IN1] = input->n[OUT1];
    n[IN2] = input->n[OUT2];
  }
  n[IN1].e = n[IN2].e = INVALIDNODE;
  loss = 1./OPT::shortckt;
  y0.f1 = value();
  y0.f0 = LINEAR;
  m0.f1 = -loss * value();
  m0.c0 = 0.;
  ev = value();
  acg = -loss * value();
  set_converged();
  assert(!constant()); /* because of incomplete analysis */
  return *this;
}
/*--------------------------------------------------------------------------*/
bool DEV_CCVS::dotr()
{
  assert(!isbypassed());
  m0.x = n[IN1].v0() - n[IN2].v0();
  y0.x = input->tr_amps();
  if (has_tr_eval()){
    tr_eval();
    m0.c0 = y0.f0 - y0.x * y0.f1;
  }else{
    assert(converged());
    y0.f1 = value();
    y0.f0 = y0.x * y0.f1;
    m0.c0 = 0.;
  }
  m0.c0 += y0.f1 * input->m0.c0;
  m0.f1  = y0.f1 * (input->loss + input->m0.f1);
  m0.c0 *= -loss;
  m0.f1 *= -loss;
  store_values();
  SIM::loadq.push(this);
  return converged();
}
/*--------------------------------------------------------------------------*/
CARD& DEV_CCVS::doac()
{
  if (!input->evaluated()){
    untested();
    input->doac();	    /* BUG: premature load of sense element */
  }
  acload_loss();
  if (has_ac_eval()){
    untested();
    acbias = n[IN1].vdc() - n[IN2].vdc();
    ac_eval();
  }else{
    ev  = y0.f1;
  }
  if (input->issource()){
    acg = -loss * ev * input->acg;
    acload_source();
    acg = -loss * ev * input->loss;
    acload_active();
  }else{
    acg = -loss * ev * input->acg;
    acload_active();
  }
  return *this;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
