/*$Id: d_trln.cc,v 13.32 1999/04/25 07:12:45 al Exp $ -*- C++ -*-
 * Transmission line. (ideal lossless.  for now, AC only)
 */
#include "ap.h"
#include "d_trln.h"
/*--------------------------------------------------------------------------*/
//		TRANSLINE_COMMON::TRANSLINE_COMMON();
//		DEV_TRANSLINE::DEV_TRANSLINE();
//	void	DEV_TRANSLINE::parse(CS *cmd);
// 	void	DEV_TRANSLINE::print(int,int)const;
// 	bool	DEV_TRANSLINE::dotr();
// 	void	DEV_TRANSLINE::doac();
/*--------------------------------------------------------------------------*/
static const double LINLENTOL = .000001;
static const double DEFAULT_NL = 0.25;
static TRANSLINE_COMMON Default_TRANSLINE(CC_STATIC);
/*--------------------------------------------------------------------------*/
TRANSLINE_COMMON::TRANSLINE_COMMON(int c)
  :COMPONENT_COMMON(c),
   z0(0),
   td(0),
   f(0),
   nl(DEFAULT_NL),
   reson(0),
   icset(false)
{
  for (int i = 0;  i < NUM_INIT_COND;  ++i){
    ic[i] = 0.;
  }
}
/*--------------------------------------------------------------------------*/
TRANSLINE_COMMON::TRANSLINE_COMMON(const TRANSLINE_COMMON& p)
  :COMPONENT_COMMON(p),
   z0(p.z0),
   td(p.td),
   f(p.f),
   nl(p.nl),
   reson(p.reson),
   icset(p.icset)
{
  untested();
  for (int i = 0;  i < NUM_INIT_COND;  ++i){
    ic[i] = p.ic[i];
  }
}
/*--------------------------------------------------------------------------*/
DEV_TRANSLINE::DEV_TRANSLINE()
{
  untested();
  attach_common(&Default_TRANSLINE);
}
/*--------------------------------------------------------------------------*/
CARD& DEV_TRANSLINE::parse(CS& cmd)
{
  assert(common);
  const TRANSLINE_COMMON* cc = dynamic_cast<const TRANSLINE_COMMON*>(common);
  assert(cc);
  TRANSLINE_COMMON* c = new TRANSLINE_COMMON(*cc);
  assert(c);

  parselabel(cmd);
  parsenodes(cmd,numnodes());
  cmd.stuck();
  do{
    untested();
    cmd.get("Z",    &c->z0, &c->z0);
    cmd.get("Freq", &c->f,  mPOSITIVE);
    cmd.get("Nl",   &c->nl, mPOSITIVE);
    if (cmd.pmatch("Ic")){
      untested();
      c->icset = true;
      for (int i = 0;  i < NUM_INIT_COND;  i++){
	untested();
	c->ic[i] = cmd.ctof();
      }
    }else{
      untested();
    }
  }while (cmd.more() && !cmd.stuck());
  cmd.check(bWARNING);

  if (c->nl == 0.){
    untested();
    c->nl = DEFAULT_NL;
  }else{
    untested();
  }
  c->reson = c->f * (.25 / c->nl);
  attach_common(c);
  return *this;
}
/*--------------------------------------------------------------------------*/
const CARD& DEV_TRANSLINE::print(OMSTREAM where, int)const
{
  const TRANSLINE_COMMON* c = dynamic_cast<const TRANSLINE_COMMON*>(common);
  assert(c);
  printlabel(where);
  printnodes(where,numnodes());
  where.setfloatwidth(7)
    << "  Z0=" << c->z0
    << "  F="  << c->f
    << "  NL=" << c->nl;
  if (c->icset){
    untested();
    where << "  IC=";
    for (int i = 0;  i < NUM_INIT_COND;  i++){
      untested();
      where << c->ic[i] << ' ';
    }
  }else{
    untested();
  }
  where << '\n';
  return *this;
}
/*--------------------------------------------------------------------------*/
/* transmission line for transient analysis
 * stub: doesn't work
 * always returns success
 */
bool DEV_TRANSLINE::dotr()
{
  untested();
  assert(!isbypassed());
  error(bWARNING, "%s: no transmission line in dc or transient\n",
	printlabel());
  set_converged();
  return converged();
}
/*--------------------------------------------------------------------------*/
CARD& DEV_TRANSLINE::doac()
{
  const TRANSLINE_COMMON* c = dynamic_cast<const TRANSLINE_COMMON*>(common);
  assert(c);
  double lenth = SIM::freq / c->reson;	/* length in quarter waves */
  double dif = lenth - floor(lenth+.5);	/* avoid divide by zero if close to */
  if (fabs(dif) < LINLENTOL){		/* resonance by tweeking a little */
    untested();
    error(bDEBUG,
	  "%s: transmission line too close to resonance\n", printlabel());
    lenth = (dif<0.) ? floor(lenth+.5)-LINLENTOL : floor(lenth+.5)+LINLENTOL;
  }else{
    untested();
  }
  lenth *= (kPId2);	/* now in radians */
  
  COMPLEX y12 = COMPLEX(0., -1. / (c->z0 * sin(lenth)));
  COMPLEX y11 = COMPLEX(0., tan(lenth/2) / c->z0) + y12;
  
  acx.s(n[OUT1].m,n[OUT1].m) += y11; 	/* BUG: bypasses load functions */
  acx.s(n[OUT2].m,n[OUT2].m) += y11;	/* result is flags may not be   */
  acx.s(n[OUT1].m,n[OUT2].m) -= y11;	/* updated.  No problem yet in  */
  acx.s(n[OUT2].m,n[OUT1].m) -= y11; 	/* AC but this will not work in */
					/* transient.			*/
  acx.s(n[IN1].m,n[IN1].m) += y11;
  acx.s(n[IN2].m,n[IN2].m) += y11;
  acx.s(n[IN1].m,n[IN2].m) -= y11;
  acx.s(n[IN2].m,n[IN1].m) -= y11;
  
  acx.s(n[OUT1].m,n[IN1].m) -= y12;
  acx.s(n[OUT2].m,n[IN2].m) -= y12;
  acx.s(n[OUT1].m,n[IN2].m) += y12;
  acx.s(n[OUT2].m,n[IN1].m) += y12;
  
  acx.s(n[IN1].m,n[OUT1].m) -= y12;
  acx.s(n[IN2].m,n[OUT2].m) -= y12;
  acx.s(n[IN1].m,n[OUT2].m) += y12;
  acx.s(n[IN2].m,n[OUT1].m) += y12;
  return *this;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
