/*$Id: e_compon.h,v 13.43 1999/06/01 15:25:51 al Exp $ -*- C++ -*-
 * branch structure type definitions
 * device types (enumeration type?)
 */
#ifndef E_COMPON_H
#define E_COMPON_H
#include "m_cpoly.h"
#include "e_model.h"
/*--------------------------------------------------------------------------*/
class COMPONENT_COMMON;
class COMPONENT;
/*--------------------------------------------------------------------------*/
#define conchk(o,n,a)	(fabs((n)-(o))<=(OPT::reltol*fabs(n)+(a)))
enum {CC_STATIC=27342383}; // mid-sized arbitrary positive int
// pass this as an argument to a common constructor to mark it as static,
// so it won't be deleted
/*--------------------------------------------------------------------------*/
class COMPONENT_COMMON {
private:
  COMPONENT_COMMON& operator=(const COMPONENT_COMMON&)
					{unreachable(); return *this;}
protected:
  explicit COMPONENT_COMMON(const COMPONENT_COMMON& p, int c=0)
    :tnom(p.tnom), _model(p._model), _attachcount(c)
    {strcpy(_modelname, p._modelname);}
  explicit COMPONENT_COMMON(int c=0)
    : tnom(NOT_INPUT), _model(0), _attachcount(c) {_modelname[0] = '\0';}
public:
  virtual ~COMPONENT_COMMON()
			{assert(_attachcount==0||_attachcount==CC_STATIC);}

  const MODEL_CARD* attach_model()const;
  COMPONENT_COMMON& attach(const MODEL_CARD* m) {_model = m; return *this;}
  COMPONENT_COMMON& set_modelname(const char* n)
					{strcpy(_modelname, n); return *this;}

  virtual void	parse(CS&)		{unreachable();}
  virtual void	print(OMSTREAM)const	{unreachable();}

  virtual void	tr_eval(COMPONENT*x)const
			{untested(); assert(_model); _model->tr_eval(x);}
  virtual void	ac_eval(COMPONENT*x)const
			{untested(); assert(_model); _model->ac_eval(x);}
  virtual bool	has_tr_eval()const	{untested(); return false;}
  virtual bool	has_ac_eval()const	{untested(); return false;}

  virtual const char* name()const	= 0;
  virtual const char* augname()const	{return name();}
	  const char* modelname()const	{return _modelname;}
    const MODEL_CARD* model()const	{assert(_model); return _model;}
	  bool	      has_model()const	{return _model;}

  const COMPONENT_COMMON* attach()const	{++_attachcount; return this;}
  int detach()const		{assert(_attachcount>0);return --_attachcount;}
protected:
  double	tnom;
private:
  char		_modelname[LABELEN+1];
  mutable const MODEL_CARD* _model;
  mutable int	_attachcount;
};
/*--------------------------------------------------------------------------*/
/* note on _attachcount ...
 * The int argument is the initial _attachcount (default = 0)
 * Set it to CC_STATIC for static default versions that will never be deleted.
 * Set it to 0 (default) for auto versions, so they can be deleted.
 * A common will not be deleted on a detach if its _attachcount != 0
 * A failed assertion from the common destructor probably means
 * the common is being deleted before a device it is attached to is,
 * without being first detached.
 * This is why ~COMPONENT destroys the subckt explicitly.
 */
/*--------------------------------------------------------------------------*/
class COMPONENT : public CARD {
protected:
  explicit COMPONENT();
  explicit COMPONENT(const COMPONENT& p);
	   ~COMPONENT()			{subckt().destroy(); detach_common();}

  void	   parselabel(CS&);
  int	   parsenodes(CS&,int,int=0,int=0);
  void	   printnodes(OMSTREAM,int,int=0)const;
  void	   tr_eval();
  void	   ac_eval();
  bool	   has_tr_eval()const
		{return (has_probes() || (common && common->has_tr_eval()));}
  bool	   has_ac_eval()const
		{return (has_probes() || (common && common->has_ac_eval()));}

  bool	   isdevice()const		{return true;}

  const MODEL_CARD* attach_model()const	{return common->attach_model();}
  void	   attach_common(const COMPONENT_COMMON*);
  void	   detach_common();

  bool	   conv_check()const	{return conchk(y1.f1, y0.f1, OPT::abstol)
				   && conchk(y1.f0, y0.f0, OPT::abstol)
				   && conchk(y1.x,  y0.x,  OPT::abstol);}
  bool	   conv_trace()const;
  bool	   converged()const		{return _converged;}
  COMPONENT& set_converged(bool s=true)	{_converged = s; return *this;}
  COMPONENT& set_not_converged()	{_converged = false;return *this;}

  void	   expandsubckt(const char* modelname);
public:
  COMPONENT& set(const char* Label, CARD* Parent,
	       const COMPONENT_COMMON* Common, double Value,
	       node_t N0, node_t N1);
  COMPONENT& set(const char* Label, CARD* Parent,
	       const COMPONENT_COMMON* Common, double Value,
	       node_t N0, node_t N1, node_t N2, node_t N3);
  CARD&	     setvalue(double v)		{CARD::setvalue(v); return *this;}
  CARD&	     setvalue(double v, const COMPONENT_COMMON* c)
					{common = c; return setvalue(v);}
public:
  const COMPONENT_COMMON* common;
public: // commons
  fpoly1_t y0;		// iteration parameters, new
  fpoly1_t y1;		// iteration parameters, 1 iter ago
  fpoly1_t y2;		// iteration parameters, 2 iter ago
  COMPLEX  ev;		// ac effective value (usually real)
  double   acbias;	// dc bias for ac use
private:
  bool _converged;	// convergence status
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
